/* Parser of crossroads configuration files */

%{
/* Prologue */
#include "crossroads.h"

#define YYSTYPE Confsets

/* Parser debugging related */
// #define PARSER_DEBUG    
#ifdef PARSER_DEBUG
    static void pmsg (char const *x) {
	printf ("P: %s\n", x);
    }
    static void psmsg (char const *x, char const *y) {
	printf ("P: %s %s\n", x, y);
    }
    static void pimsg(char const *x, int y) {
	printf ("P: %s %d\n", x, y);
    }
#else
#   define pmsg(x)
#   define psmsg(x,y)
#   define pimsg(x,y)
#endif

/* Error handler for yyparse() */    
static int yyerror (char *msg) {
    error ("Parse error at line %d, '%s': %s",
	   yylineno + 1, yytext, yyerrmsg);
}

/* Store an encountered string */    
static char *laststr;
static void setlaststr (char const *what) {
    free (laststr);
    laststr = xstrdup (what);
}

/* Store an encountered number */
static int lastnr;
static void setlastnr (char const *what) {
    lastnr = atoi (what);
}

/* Store an encountered 'over' number */
static int lastovernr;
static void setlastovernr (char const *what) {
    lastovernr = atoi(what);
}

/* Temp vars */
static int i;				/* Loop counter */
static Backend cur_backend;		/* Storage for a handled backend */
static Service cur_service;		/* Storage for a handled service */
%}

/* Declarations */
%token SERVICE IDENTIFIER PORT NUMBER BACKEND VERBOSITY SERVER
       ON OFF DISPATCHMODE ROUNDROBIN REVIVINGINTERVAL SHMKEY WEIGHT
       ONSUCCESS ONFAILURE STRING BACKLOG RANDOM BYDURATION BYSIZE
       SESSIONTIMEOUT MAXCLIENTS BYORDER DUMPTRAFFIC OVER DECAY BINDTO

%%
/* Config file grammar rules */

input:
	element
	input
|
	element
;

element:
	service
	servicename
	'{'
	servicestatements
	'}' {
		/* Verify the service description, supply defaults
		 * and so on.
		 */
	        if (!cur_service.port)
		    error ("Service %s lacks a port",
			   cur_service.name);
		if (!cur_service.nbackend)
		    error ("Service %s lacks back ends",
			   cur_service.name);

		if (!cur_service.shmkey)
		    cur_service.shmkey = cur_service.port | SHM_MASK;

		if (!cur_service.bind)
		    cur_service.bind = "any";

		/* Add to the list. */
	    	service = xrealloc (service, ++nservice * sizeof(Service));
		service[nservice - 1] = cur_service;
		memset (&cur_service, 0, sizeof(Service));
	}
;

service:
	service_expected
	SERVICE
;

servicename:
	servicename_expected
	IDENTIFIER {
		psmsg ("service:", yytext);
		cur_service.name = xstrdup(yytext);
	}
;

servicestatements:
	servicestatements
	servicestatement
|
	servicestatement
;

servicestatement:
	servicebody_expected
	servicebody
;

servicebody:
	portstatement {
	        pimsg ("sevice port:", $1.set[0].v.ival);
		cur_service.port = $1.set[0].v.ival;
		free ($1.set);
	}
|	
	bindstatement {
		psmsg ("service binding:", $1.set[0].v.sval);
		cur_service.bind = $1.set[0].v.sval;
		free ($1.set);
	}
|
	verbositystatement {
	        pimsg ("service verbosity:", $1.set[0].v.ival);
		cur_service.verbosity = $1.set[0].v.ival;
		free ($1.set);
	}
|
	dispatchmodestatement {
	        pimsg ("service dispatch mode:", $1.set[0].v.ival);
		pimsg ("service dispatch over:", lastovernr);
		cur_service.dispatchtype = $1.set[0].v.ival;
		cur_service.dispatchover = lastovernr;
		free ($1.set);
	}
|
	revivingintervalstatement {
	        pimsg ("service revival interval:", $1.set[0].v.ival);
		cur_service.rev_interval = $1.set[0].v.ival;
		free ($1.set);
	}
|
	backlogstatement {
		pimsg ("service backlog:", $1.set[0].v.ival);
		cur_service.backlog = $1.set[0].v.ival;
		free ($1.set);
	}
|
	shmkeystatement {
		pimsg ("service shmkey:", $1.set[0].v.ival);
		cur_service.shmkey = $1.set[0].v.ival;
		free ($1.set);
	}
|
	sessiontimeoutstatement {
	    	pimsg ("session timout:", $1.set[0].v.ival);
		cur_service.sessiontimeout = $1.set[0].v.ival;
		free ($1.set);
	}
|
	maxclientsstatement {
		pimsg ("max clients:", $1.set[0].v.ival);
		cur_service.maxclients = $1.set[0].v.ival;
		free ($1.set);
	}
|
	backendstatement {
 	        pimsg ("converting backend statements, count is", $1.n);
		for (i = 0; i < $1.n; i++)
		    switch ($1.set[i].cf) {
		    case cf_portspec:
			pimsg ("backend block port:", $1.set[i].v.ival);
			cur_backend.port = $1.set[i].v.ival;
			break;
		    case cf_serverspec:
			psmsg ("backend block server:", $1.set[i].v.sval);
			cur_backend.server = $1.set[i].v.sval;
			break;
		    case cf_verbosityspec:
			pimsg ("backend block verbosity:", $1.set[i].v.ival);
			cur_backend.verbosity = $1.set[i].v.ival;
			break;
		    case cf_onsuccspec:
			psmsg ("backend block onsuccess:", $1.set[i].v.sval);
			cur_backend.onsuccess = $1.set[i].v.sval;
			break;
		    case cf_onfailspec:
			psmsg ("backend block onfailure:", $1.set[i].v.sval);
			cur_backend.onfailure = $1.set[i].v.sval;
			break;
		    case cf_dumpspec:
			psmsg ("backend dumptraffic:", $1.set[i].v.sval);
			cur_backend.dumpfile = $1.set[i].v.sval;
			break;
		    case cf_weightspec:
			pimsg ("backend weight:", $1.set[i].v.ival);
			cur_backend.weight = $1.set[i].v.ival;
			break;
		    case cf_decayspec:
			pimsg ("backend decay:", $1.set[i].v.ival);
			if ($1.set[i].v.ival >= 100)
			    error ("Decay specifier %d must be a percentage, "
				   "never more than 99",
				   $1.set[i].v.ival);
			cur_backend.decay = $1.set[i].v.ival;
			break;
		    default:
			error ("Internal jam, unhandled type %d "
			       "in backend specification",
			       $1.set[i].cf);
		    }
		free ($1.set);
		/* Verify the backend block, supply defaults,
		 * And so on.
		 */
		if (!cur_service.port)
		    error ("Back end %s lacks port",
			   cur_service.port);
		if (!cur_backend.weight)
		    cur_backend.weight = 1;

		/* Add to the list. */
		cur_service.backend = xrealloc (cur_service.backend,
						++cur_service.nbackend *
						sizeof(Service));
		cur_service.backend[cur_service.nbackend - 1] =
		    cur_backend;
		pimsg ("this was backend defintion", cur_service.nbackend);
		memset (&cur_backend, 0, sizeof(cur_backend));
	}
;

portstatement:
	PORT
	number
	semicol {
	        pimsg ("port statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_portspec;
		$$.set[0].v.ival = lastnr;
	}
;

bindstatement:
	BINDTO
	ipaddress
	semicol {
	        psmsg ("bindto statement:", laststr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_bindspec;
		$$.set[0].v.sval = xstrdup(laststr);
	}
;

ipaddress:
	ipaddress_expected
	STRING {
	       setlaststr (laststring);
	       free (laststring);
	       laststring = 0;
	}
;

number:
	number_expected
	NUMBER {
		setlastnr (yytext);
	}
;

semicol:
	semicol_expected
	';'
;

verbositystatement:
	VERBOSITY
	onoff_expected
	onoff
	semicol {
	        pimsg ("verbosity statement:", lastnr);
	    	$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_verbosityspec;
		$$.set[0].v.ival = lastnr;
	}
;

onoff:
	ON {
		lastnr = 1;
	}
|
	OFF {
		lastnr = 0;
	}
;

dispatchmodestatement:
	DISPATCHMODE
	dispatchmethod
	opt_over
	semicol {
		pimsg ("dispatch mode statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_dispatchspec;
		$$.set[0].v.ival = lastnr;
	}
;

opt_over:
	OVER
	overnumber
|
	/* empty */ {
		lastovernr = 0;
	}
;

overnumber:
	number_expected
	NUMBER {
		setlastovernr (yytext);
	}
;

dispatchmethod:
	dispatchmethod_expected
	dispatchmethodspec
;

dispatchmethodspec:
	ROUNDROBIN {
		lastnr = ds_roundrobin;
	}
|
	RANDOM {
	       lastnr = ds_random;
	}
|
	BYDURATION {
	    	lastnr = ds_byduration;
	}
|
	BYSIZE {
		lastnr = ds_bysize;
	}
|
	BYORDER {
		lastnr = ds_byorder;
	}
;

revivingintervalstatement:
	REVIVINGINTERVAL
	number
	semicol {
		pimsg ("reviving interval statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_revivespec;
		$$.set[0].v.ival = lastnr;
	}
;

backlogstatement:
	BACKLOG
	number
	semicol {
		pimsg ("backlog statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_revivespec;
		$$.set[0].v.ival = lastnr;
	}
;

shmkeystatement:
	SHMKEY
	number
	semicol {
		pimsg ("shmkey statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_shmkeyspec;
		$$.set[0].v.ival = lastnr;
	}
;

sessiontimeoutstatement:
	SESSIONTIMEOUT
	number
	semicol {
		pimsg ("session timeout statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_sessiontimeoutspec;
		$$.set[0].v.ival = lastnr;
	}
;

maxclientsstatement:
	MAXCLIENTS
	number
	semicol {
		pimsg ("max clients statement:", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof(Confset));
		$$.set[0].cf = cf_maxclientsspec;
		$$.set[0].v.ival = lastnr;
	}
;

backendstatement:
	BACKEND
	backendname
	backendblock {
		$$ = $3;
	}
;

backendname:
	backendname_expected
	IDENTIFIER {
		psmsg ("backend name:", laststr);
		cur_backend.name = xstrdup (yytext);
	}
;

backendblock:
	'{'
	backenddefinitions
	'}' {
		$$ = $2;
	}
;

backenddefinitions:
	backenddefinitions
	backenddefinition {
		$1.n++;
		$1.set = xrealloc ($1.set, $1.n * sizeof(Confset));
		$1.set[$1.n - 1] = $2.set[0];
		$$ = $1;
	}
|
	backenddefinition {
		$$ = $1;
	}
;

backenddefinition:
	backendstatement_expected
	backendstatement {
		$$ = $2;
	}
;

backendstatement:
	serverstatement {
		psmsg ("backend server:", $1.set[0].v.sval);
		$$ = $1;
	}
|
	portstatement {
		pimsg ("backend port:", $1.set[0].v.ival);
		$$ = $1;
	}
|
	verbositystatement {
		pimsg ("backend verbosity:", $1.set[0].v.ival);
		$$ = $1;
	}
|
	onsuccessstatement {
		psmsg ("backend onsuccess:", $1.set[0].v.sval);
		$$ = $1;
	}
|
	onfailurestatement {
		psmsg ("backend onfailure:", $1.set[0].v.sval);
		$$ = $1;
	}
|
	dumptrafficstatement {
		psmsg ("backend dumptraffic:", $1.set[0].v.sval);
		$$ = $1;
	}
|
	weightstatement {
	    	pimsg ("backend weight:", $1.set[0].v.ival);
		$$ = $1;
	}
|
	decaystatement {
	    	pimsg ("backend decay:", $1.set[0].v.ival);
		$$ = $1;
	}
;

serverstatement:
	SERVER
	serveraddress_expected
	serveraddress
	semicol {
		psmsg ("server statement:", laststr);
		$$.n = 1;
		$$.set = xmalloc (sizeof (Confset));
		$$.set[0].cf = cf_serverspec;
		$$.set[0].v.sval = xstrdup (laststr);
	}
;

weightstatement:
	WEIGHT
	number
	semicol {
		pimsg ("weight statement", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof (Confset));
		$$.set[0].cf = cf_weightspec;
		$$.set[0].v.ival = lastnr;
	}
;

decaystatement:
	DECAY
	number
	semicol {
		pimsg ("decay statement", lastnr);
		$$.n = 1;
		$$.set = xmalloc (sizeof (Confset));
		$$.set[0].cf = cf_decayspec;
		$$.set[0].v.ival = lastnr;
	}
;

serveraddress:
	STRING {
		setlaststr (laststring);
	}
;

onsuccessstatement:
	ONSUCCESS
	commandline
	semicol {
		psmsg ("onsuccess statement:", laststr);
		$$.n = 1;
		$$.set = xmalloc (sizeof (Confset));
		$$.set[0].cf = cf_onsuccspec;
		$$.set[0].v.sval = xstrdup (laststr);
	}
;

onfailurestatement:
	ONFAILURE
	commandline
	semicol {
		psmsg ("onfailure statement:", laststr);
		$$.n = 1;
		$$.set = xmalloc (sizeof (Confset));
		$$.set[0].cf = cf_onfailspec;
		$$.set[0].v.sval = xstrdup (laststr);
	}
;

dumptrafficstatement:
	DUMPTRAFFIC
	filename
	semicol {
		psmsg ("dumptraffic statement:", laststr);
		$$.n = 1;
		$$.set = xmalloc (sizeof (Confset));
		$$.set[0].cf = cf_dumpspec;
		$$.set[0].v.sval = xstrdup (laststr);
	}
;

commandline:
	commandline_expected
	STRING {
		setlaststr (laststring);
		free (laststring);
		laststring = 0;
	}
;

filename:
	filename_expected
	STRING {
		setlaststr (laststring);
		free (laststring);
		laststring = 0;
	}
;

number_expected: {
		yyerrmsg = "number expected";
	}
;

serveraddress_expected: {
		yyerrmsg = "hostname or IP address expected";
	}
;

service_expected: {
		yyerrmsg = "'service' expected";
	}
;

backendstatement_expected: {
		yyerrmsg = "backend definition statement expected";
	}
;

servicebody_expected: {
		yyerrmsg = "service body statement expected";
	}
;

semicol_expected: {
		yyerrmsg = "semicolon (;) expected";
	}
;

onoff_expected: {
    		yyerrmsg = "'on' or 'off' expetcted";
	}
;

dispatchmethod_expected: {
		yyerrmsg = "dispatch method expected";
	}
;

commandline_expected: {
		yyerrmsg = "command line expected";
	}
;

filename_expected: {
		yyerrmsg = "file name expected";
	}
;

servicename_expected: {
		yyerrmsg = "service name (identifier) expected";
	}
;

backendname_expected: {
		yyerrmsg = "backend name (identifier) expected";
	}
;

ipaddress_expected: {
		yyerrmsg = "IP address or 'any' expected";
	}
;
